#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>

#include "nokia_3310_lcd.h"
#include "virtual_timer.h"

/* Graphical symbols */
#include "diode.h"
#include "bt_power.h"
#include "bt_menu.h"
#include "bt_plus.h"
#include "bt_minus.h"

/* I/O pins definition*/
#define BUTTON_PLUS_PIN PD0
#define BUTTON_MINUS_PIN PD1
#define BUTTON_MENU_PIN PD2

#define PWM_OUT PB1
#define AUTO_OFF PD6
#define PIEZZO PD3
#define BACKLIGHT_CONTROL PD5
#define OUTPUT_ON_OFF PC2

/* States */
#define BUTTON_PLUS 0
#define BUTTON_MINUS 1
#define BUTTON_MENU 2

#define BUTTON_NONE 0
#define BUTTON_PRESS 1
#define BUTTON_HOLD 2
#define BUTTON_RELEASE 3

#define SIMPLE_VIEW 0
#define ADVANCED_VIEW 1
#define SETTINGS_VIEW 2
#define CALIBRATION_VIEW 3

#define CALIBRATION_1 0
#define CALIBRATION_2 1

/* Constants */
#define NUMBER_OF_BUTTONS 3
#define SAMPLES 256
#define BEEP_LENGHT 20
#define MAX_LED_CURRENT 1000
#define MIN_LED_CURRENT 0
#define MAX_LED_POWER_SUPPLY 3000
#define MIN_LED_POWER_SUPPLY 0
#define LED_CURRENT_STEP1 1
#define LED_CURRENT_STEP2 5
#define LED_POWER_SUPPLY_STEP1 10

#define VTIMER_0 0		// Display refresh
#define VT0_PERIOD 1000	// 20Hz

#define VTIMER_1 1		// Buttons
#define VT1_PERIOD 40	// 200Hz

#define VTIMER_2 2		// Control loop
#define VT2_PERIOD 500	// 20Hz

#define VTIMER_3 3		// Auto Power Off
#define VT3_PERIOD 10000

#define VTIMER_4 4		// Beep frequency
#define VT4_PERIOD 10

/* Variables*/

volatile uint8_t active_view = SIMPLE_VIEW;

uint8_t button_array[3]={BUTTON_PLUS_PIN, BUTTON_MINUS_PIN, BUTTON_MENU_PIN};

volatile uint8_t button_on_counter[NUMBER_OF_BUTTONS];
volatile uint8_t button_off_counter[NUMBER_OF_BUTTONS];
volatile uint8_t buttons_status[NUMBER_OF_BUTTONS];

uint8_t hold_button_counter = 0;
uint8_t previous_view;
uint8_t back_from_settings = 0;
uint8_t to_settings = 0;
uint8_t led_power_supply_change = 0;
volatile uint8_t update_flag = 0;
volatile uint8_t display_update = 1;
uint8_t calibration_step = 0;

uint16_t auto_power_off_counter = 0;
uint16_t auto_power_off_counter_max = 60;
uint8_t beep_lenght_counter = 0;
uint8_t speaker_flag;

uint16_t cal_const1 = 790;
uint16_t cal_const2 = 6280;

int32_t current_control_const_a = 6000;
int32_t current_control_const_b = 0;

volatile uint16_t LED_current = 0;			// mA*10
volatile uint16_t LED_voltage = 0;			// V*100
uint16_t LED_power_supply = 0;				// V*100
int16_t LED_resistor = 0;					// ohms*100
uint8_t LED_resistor_unit = 0;				// 0..-, 1..k, 2..M
volatile uint16_t LED_current_meas = 0;

uint32_t LED_current_raw = 0;
uint32_t LED_voltage_raw = 0;
uint16_t voltage_sample_index = 0;
uint16_t current_sample_index = 0;

uint8_t config_backlight = 1;
uint8_t config_beep = 0;
uint8_t config_resistor = 1;
uint8_t config_auto_off = 1;

uint8_t active_ADC_channel = 0;

volatile uint8_t menu_item_selected = 0;
volatile uint8_t menu_item_cursor = 0;

uint32_t temp_value = 0;

ISR(TIMER0_COMPA_vect) 
{
	vt_interrupt();
}

ISR(ADC_vect)
{
	switch (active_ADC_channel)
	{
		case 0:
			LED_voltage_raw += ADC;
			
			ADMUX |= (1<<MUX0);
			active_ADC_channel = 1;
			
			voltage_sample_index += 1;
	
			if (voltage_sample_index == SAMPLES) 
			{
				temp_value = LED_voltage_raw / SAMPLES;
				LED_voltage = ((uint32_t)1434*(uint32_t)temp_value)/(uint32_t)1024; // 3.3 * 100 / 0.230 = 1434 reference voltage * 100 / gain
				LED_voltage_raw = 0;
				voltage_sample_index = 0;
				display_update = 1;
			}
		break;
	
		case 1:
			LED_current_raw += ADC;
			
			ADMUX &= ~(1<<MUX0);
			active_ADC_channel = 0;
			
			current_sample_index += 1;
	
			if (current_sample_index == SAMPLES) 
			{
				temp_value = LED_current_raw / SAMPLES;
				LED_current_meas = ((uint32_t)1100*(uint32_t)temp_value)/((uint32_t)1024); // 3.3 / 30 * 10 * 1000 = 1100 reference voltage / gain * 10 * mA
				LED_voltage -= LED_current_meas/100; 
				LED_current_raw = 0;
				current_sample_index = 0;
				display_update = 1;
			}
		break;
		
		case 2:
		break;
	};
	
	ADCSRA |= (1<<ADSC);
}

void peripherals_init()
{
	// Configure pin as output for power on switch
	DDRD |= (1<<AUTO_OFF);
	
	// Hold power on switch
	PORTD |= (1<<AUTO_OFF);
	
	// Timer0 Initialization
	TCCR0A |= (1<<WGM01);					// Set timer0 on compare match (CTC) mode
	OCR0A = 25;								// Set value of compare match register, interrupt 10kHz
	TIMSK0 |= (1<<OCIE0A);					// Enable compare match A interrupt enable
	TCCR0B |= (1<<CS01) | (1<<CS00);		// Start clock 1/64 prescaler
	
	// Configure I/O pins
	// Set Pull-up rezistors for buttons
	PORTD |= (1<<BUTTON_PLUS_PIN) | (1<<BUTTON_MINUS_PIN) | (1<<BUTTON_MENU_PIN);
	// Configure pin as output for piezzo and backlight_control
	DDRD |= (1<<PIEZZO) | (1<<BACKLIGHT_CONTROL);
	//
	DDRC |= (1<<OUTPUT_ON_OFF);
	PORTC &= ~(1<<OUTPUT_ON_OFF);
	
	// PWM generator init
	DDRB |= (1<<PWM_OUT);					// Configure PWM pin as output
	TCCR1A |= (1<<COM1A1);					// Clear on compare match, set at bottom
	TCCR1A |= (1<<WGM11);					// Fast PWM mode 16bit
	TCCR1B |= (1<<WGM12) | (1<<WGM13);		// Fast PWM mode 16bit
	
	ICR1 = 0xFFFF;							// Set the top to maximum
	OCR1A = 100;							// Initial value
	
	TCCR1B |= (1<<CS10);					// Start counting, no prescaller
	
	// ADC init
	ADMUX |= (1<<REFS0);					// Aref = AVcc with external blocking capacitor
	ADCSRA |= (1<<ADEN) | (1<<ADIE);		// ADC enable, Interrupt enable
	ADCSRA |= (1<<ADPS1) | (1<<ADPS2);		// ADC clock pre-scaler 64	
	ADCSRA |= (1<<ADSC);					// ADC start
}

void show_welcome_screen()
{
	nokia_3310_lcd_writeString(8,0,"DIODE TESTER", MENU_NORMAL);
	nokia_3310_lcd_writeString(8,5,"v1.0    (MB)", MENU_NORMAL);
	nokia_3310_lcd_drawBitmapP(25,1,&diode_image[0],32,32);

	_delay_ms(2000);
}

void refresh_display()
{
	vt_clear_flag(VTIMER_0);
	
	if (display_update)
	{
		char cTemp[6];
		
		switch(active_view)
		{
			case SIMPLE_VIEW:
				nokia_3310_lcd_clear();
				
				nokia_3310_lcd_writeString(0,0,"Iled:", MENU_NORMAL);
				nokia_3310_lcd_writeString(0,1,"[mA]", MENU_NORMAL);
				sprintf(cTemp,"%.3d.%.1d", LED_current/10, LED_current%10);
				nokia_3310_lcd_writeStringBig(28,0,cTemp, MENU_NORMAL);
								
				nokia_3310_lcd_writeString(0,3,"Vled:", MENU_NORMAL);
				nokia_3310_lcd_writeString(0,4,"[V]", MENU_NORMAL);
				sprintf(cTemp,"%.2d.%.2d", LED_voltage/100, LED_voltage%100);
				nokia_3310_lcd_writeStringBig(28,6,cTemp, MENU_NORMAL);
			break;
		
			case ADVANCED_VIEW:
				nokia_3310_lcd_clear();
				
				nokia_3310_lcd_writeString(0,1,"Iled:", MENU_NORMAL);
				sprintf(cTemp,"%.3d.%.1d", LED_current/10, LED_current%10);
				nokia_3310_lcd_writeString(35,1,cTemp, MENU_NORMAL);
				nokia_3310_lcd_writeString(70,1,"mA", MENU_NORMAL);
				
				nokia_3310_lcd_writeString(0,2,"Uled:", MENU_NORMAL);
				sprintf(cTemp,"%.2d.%.2d", LED_voltage/100, LED_voltage%100);
				nokia_3310_lcd_writeString(35,2,cTemp, MENU_NORMAL);
				nokia_3310_lcd_writeString(70,2,"V", MENU_NORMAL);
				
				nokia_3310_lcd_writeString(0,3,"Vcc:", MENU_NORMAL);
				sprintf(cTemp,"%.2d.%.2d", LED_power_supply/100, LED_power_supply%100);
				nokia_3310_lcd_writeString(35,3,cTemp, MENU_NORMAL);
				nokia_3310_lcd_writeString(70,3,"V", MENU_NORMAL);
				
				nokia_3310_lcd_writeString(0,4,"Rs:", MENU_NORMAL);
				
				switch (LED_resistor_unit%10)
				{
					case 0:
						sprintf(cTemp," %.1d.%.2d", LED_resistor/100, LED_resistor%100);
					break;
					
					case 1:
						sprintf(cTemp," %.2d.%.1d", LED_resistor/10, LED_resistor%10);
					break;
					
					case 2:
						sprintf(cTemp,"  %.3d", LED_resistor);	
					break;
				}
				
				nokia_3310_lcd_writeString(35,4,cTemp, MENU_NORMAL);
				
				
				switch (LED_resistor_unit/10)
				{
					case 0: nokia_3310_lcd_writeString(70,4,"^ ", MENU_NORMAL); break;
					case 1:	nokia_3310_lcd_writeString(70,4,"k^", MENU_NORMAL); break;
					case 2: nokia_3310_lcd_writeString(70,4,"M^", MENU_NORMAL); break;
				}
				
				
				nokia_3310_lcd_drawBitmapP(1,6,&bt_power_image[0],13,8);
				nokia_3310_lcd_drawBitmapP(24,6,&bt_menu_image[0],13,8);
				nokia_3310_lcd_drawBitmapP(47,6,&bt_minus_image[0],13,8);
				nokia_3310_lcd_drawBitmapP(70,6,&bt_plus_image[0],13,8);
			break;
		
			case SETTINGS_VIEW:
				nokia_3310_lcd_clear();
				
				if (menu_item_selected == 0)
					nokia_3310_lcd_writeString(0,menu_item_cursor+1,">", MENU_NORMAL);
				else
					nokia_3310_lcd_writeString(0,menu_item_cursor+1,"`", MENU_NORMAL);
				
				nokia_3310_lcd_writeString(0,0,"   SETTINGS   ", MENU_HIGHLIGHT);
				
				nokia_3310_lcd_writeString(7,1,"Backlight", MENU_NORMAL);
				switch (config_backlight)
				{
					case 0: nokia_3310_lcd_writeString(65,1,"Off", MENU_NORMAL); break;
					case 1: nokia_3310_lcd_writeString(65,1,"On", MENU_NORMAL); break;
				}
							
				nokia_3310_lcd_writeString(7,2,"Beep", MENU_NORMAL);
				switch (config_beep)
				{
					case 0: nokia_3310_lcd_writeString(65,2,"0ff", MENU_NORMAL); break;
					case 1: nokia_3310_lcd_writeString(65,2,"On", MENU_NORMAL);  break;
				}
				
				nokia_3310_lcd_writeString(7,3,"Auto-Off", MENU_NORMAL);
				switch (config_auto_off)
				{
					case 0: nokia_3310_lcd_writeString(65,3,"30s", MENU_NORMAL); break;
					case 1: nokia_3310_lcd_writeString(65,3,"01m", MENU_NORMAL); break;
					case 2: nokia_3310_lcd_writeString(65,3,"02m", MENU_NORMAL); break;
					case 3: nokia_3310_lcd_writeString(65,3,"05m", MENU_NORMAL); break;
					case 4: nokia_3310_lcd_writeString(65,3,"10m", MENU_NORMAL); break;
				}
				
				nokia_3310_lcd_writeString(7,4,"Resistor", MENU_NORMAL);
				switch (config_resistor)
				{
					case 0:	nokia_3310_lcd_writeString(65,4,"E6 ", MENU_NORMAL); break;
					case 1:	nokia_3310_lcd_writeString(65,4,"E12", MENU_NORMAL); break;	
					case 2:	nokia_3310_lcd_writeString(65,4,"E24", MENU_NORMAL); break;
					case 3:	nokia_3310_lcd_writeString(65,4,"E96", MENU_NORMAL); break;
				}
				
				nokia_3310_lcd_writeString(7,5,"Calibration", MENU_NORMAL);	
			break;
			
			case CALIBRATION_VIEW:
				nokia_3310_lcd_clear();
				
				nokia_3310_lcd_writeString(0,0," CALIBRATION  ", MENU_HIGHLIGHT);
				
				nokia_3310_lcd_writeString(11,2,"1mA ... ", MENU_NORMAL);
				sprintf(cTemp,"%.5d", cal_const1);
				nokia_3310_lcd_writeString(55,2,cTemp, MENU_NORMAL);
				
				nokia_3310_lcd_writeString(5,3,"10mA ... ", MENU_NORMAL);
				sprintf(cTemp,"%.5d", cal_const2);
				nokia_3310_lcd_writeString(55,3,cTemp, MENU_NORMAL);
				
				if (calibration_step == CALIBRATION_1)
				{
					nokia_3310_lcd_writeString(0,2,"`", MENU_NORMAL);
				}
				
				if (calibration_step == CALIBRATION_2)
				{
					nokia_3310_lcd_writeString(0,3,"`", MENU_NORMAL);
				}
				
				nokia_3310_lcd_drawBitmapP(24,6,&bt_menu_image[0],13,8);
				nokia_3310_lcd_drawBitmapP(47,6,&bt_minus_image[0],13,8);
				nokia_3310_lcd_drawBitmapP(70,6,&bt_plus_image[0],13,8);
			break;
		}
		
		display_update = 0;
	}	
	
	if (active_view != CALIBRATION_VIEW)
	{
		OCR1A = (LED_current*current_control_const_a + current_control_const_b)/(uint32_t)100;
		if (LED_current == 0) PORTC |= (1<<OUTPUT_ON_OFF); else PORTC &= ~(1<<OUTPUT_ON_OFF);
	}		
	else
	{
		switch (calibration_step)
		{
			case CALIBRATION_1: OCR1A = cal_const1;	break;
			case CALIBRATION_2: OCR1A = cal_const2;	break;
		}
		PORTC &= ~(1<<OUTPUT_ON_OFF);
	}
	
	vt_start(VTIMER_0, VT0_PERIOD);
}

void button_scan()
{
	vt_clear_flag(VTIMER_1);
	
	for (uint8_t i=0; i<NUMBER_OF_BUTTONS; i++)
	{
		if( ~PIND & (1<<button_array[i]) )
		{
			button_on_counter[i]++;
			if (button_on_counter[i]>5)
			{
				if (update_flag == 0)
				{
					switch (buttons_status[i])
					{
						case BUTTON_NONE: 
							buttons_status[i] = BUTTON_PRESS;
							beep();
							reset_auto_off_counter();
							update_flag = 1;
						break;
						case BUTTON_PRESS: 
							buttons_status[i] = BUTTON_HOLD;
							reset_auto_off_counter();
							update_flag = 1;
						break;
					}
				}
				button_on_counter[i] = 0;
				button_off_counter[i] = 0;				
			}
  		}
		else
		{
			button_off_counter[i]++;
			if (button_off_counter[i]>5)
			{
				if (update_flag == 0)
				{
					switch (buttons_status[i])
					{
						case BUTTON_HOLD:
						case BUTTON_PRESS:
							buttons_status[i] = BUTTON_RELEASE;
							reset_auto_off_counter();
							update_flag = 1;
						break;
						case BUTTON_RELEASE: 
							buttons_status[i] = BUTTON_NONE;
							reset_auto_off_counter();
							update_flag = 1;
						break;
					}
					
				}
				button_on_counter[i] = 0;
				button_off_counter[i] = 0;				
			}
		}
	}

	vt_start(VTIMER_1, VT1_PERIOD);
}

void control_loop()
{
		vt_clear_flag(VTIMER_2);
		
		switch (active_view)
		{
			case SIMPLE_VIEW:
			case ADVANCED_VIEW:
				
				resistor_update();
				
				switch (buttons_status[BUTTON_PLUS])
				{
					case BUTTON_PRESS:
						
						if (buttons_status[BUTTON_MENU] == BUTTON_HOLD)
						{
							if (LED_power_supply <= MIN_LED_POWER_SUPPLY-LED_POWER_SUPPLY_STEP1) LED_power_supply += LED_POWER_SUPPLY_STEP1;
							led_power_supply_change = 1;
							hold_button_counter = 0;						
						}
						else
						{
							if (LED_current <= MAX_LED_CURRENT-LED_CURRENT_STEP1) LED_current += LED_CURRENT_STEP1;
						}
						
						display_update = 1;
					break;
						
					case BUTTON_HOLD:
						if (buttons_status[BUTTON_MENU] == BUTTON_HOLD)
						{
							hold_button_counter = 0;	
						}
						else
						{
							if (hold_button_counter >= 20)
							{
								if (LED_current <= MAX_LED_CURRENT-LED_CURRENT_STEP1) LED_current += LED_CURRENT_STEP1; else LED_current = MAX_LED_CURRENT;
								display_update = 1;
								hold_button_counter+=1;
							}
							
							if (hold_button_counter >=50)
							{
								if (LED_current <= MAX_LED_CURRENT-LED_CURRENT_STEP2) LED_current += LED_CURRENT_STEP2; else LED_current = MAX_LED_CURRENT;
								display_update = 1;
							}
							
							if (hold_button_counter < 50) hold_button_counter += 1;							
						}							
					break;
						
					case BUTTON_RELEASE:
						hold_button_counter = 0;
					break;
				}
					
				switch (buttons_status[BUTTON_MINUS])
				{
					case BUTTON_PRESS:
						if (buttons_status[BUTTON_MENU] == BUTTON_HOLD)
						{
							if (LED_power_supply >= MIN_LED_POWER_SUPPLY+LED_POWER_SUPPLY_STEP1) LED_power_supply -= LED_POWER_SUPPLY_STEP1;
							led_power_supply_change = 1;
							hold_button_counter = 0;
						}
						else
						{
							if (LED_current >= MIN_LED_CURRENT+LED_CURRENT_STEP1) LED_current -= LED_CURRENT_STEP1;
						}							
						display_update = 1;
					break;
							
					case BUTTON_HOLD:
						if (buttons_status[BUTTON_MENU] == BUTTON_HOLD)
						{
							hold_button_counter = 0;	
						}
						else
						{
							if (hold_button_counter >= 20)
							{
								if (LED_current >= MIN_LED_CURRENT+LED_CURRENT_STEP1) LED_current -= LED_CURRENT_STEP1; else LED_current = MIN_LED_CURRENT;	
								display_update = 1;
							}
							
								if (hold_button_counter >= 50)
							{
								if (LED_current >= MIN_LED_CURRENT+LED_CURRENT_STEP2) LED_current -= LED_CURRENT_STEP2; else LED_current = MIN_LED_CURRENT;	
								display_update = 1;
							}
							
							if (hold_button_counter < 50) hold_button_counter+=1;
						}							
					break;
						
					case BUTTON_RELEASE:
						hold_button_counter = 0;
					break;
				}
					
				switch (buttons_status[BUTTON_MENU])
				{ 
					case BUTTON_PRESS:
						
					break;
						
					case BUTTON_HOLD:
						if (hold_button_counter >= 20)
						{
							previous_view = active_view;
							active_view = SETTINGS_VIEW;
							menu_item_cursor = 0;
							display_update = 1;
							hold_button_counter = 0;
							to_settings = 1;
						}
							
						if (hold_button_counter < 20) hold_button_counter+=1;								
					break;
						
					case BUTTON_RELEASE:
						if (hold_button_counter < 20)
						{
							if (back_from_settings == 0)
							{
								if (led_power_supply_change == 0)
								{
									if ((active_view == SIMPLE_VIEW))  active_view = ADVANCED_VIEW; else active_view = SIMPLE_VIEW;	
									display_update = 1;
								}
								else
								{
									led_power_supply_change = 0;
								}																			
							}
							else
							{
								back_from_settings = 0;
							}
						}
						hold_button_counter = 0;
						
					break;
				}
					
				update_flag = 0;					
			break;
			
			case SETTINGS_VIEW:
				switch (buttons_status[BUTTON_PLUS])
				{
					case BUTTON_PRESS:
						switch (menu_item_selected)
						{
							case 0:
								if (menu_item_cursor<4) menu_item_cursor+= 1; else menu_item_cursor = 0;
								display_update = 1;
							break;
							
							case 1:	
								if (config_backlight == 0) config_backlight = 1; else config_backlight = 0;
								menu_backlight_control(); 
							break;
							
							case 2:	
								if (config_beep == 0) config_beep = 1; else config_beep = 0;
							break;
							
							case 3: 
								if (config_auto_off < 4) config_auto_off++; else config_auto_off = 0;
								menu_auto_off_control(); 
							break;
							
							case 4:
								if (config_resistor < 3) config_resistor++; else config_resistor = 0;
							break;
						}
						
						display_update = 1;
					break;
				}
					
				switch (buttons_status[BUTTON_MINUS])
				{
					case BUTTON_PRESS:
						switch (menu_item_selected)
						{
							case 0:
								if (menu_item_cursor>0) menu_item_cursor-= 1; else menu_item_cursor = 4;
								display_update = 1;
							break;
							
							case 1:	
								if (config_backlight == 0) config_backlight = 1; else config_backlight = 0;
								menu_backlight_control();
							break;
							
							case 2:	
								if (config_beep == 0) config_beep = 1; else config_beep = 0;
							break;
							
							case 3: 
								if (config_auto_off > 0) config_auto_off--; else config_auto_off = 4;
								menu_auto_off_control(); 
							break;
							
							case 4:
								if (config_resistor > 0) config_resistor--; else config_resistor = 3;
							break;
						}
						
						display_update = 1;
												
					break;
				}
				
				switch (buttons_status[BUTTON_MENU])
				{	
					case BUTTON_PRESS:
					break;
					
					case BUTTON_HOLD:
						if (hold_button_counter >= 20)
							{
								write_config_eeprom();
								active_view = previous_view;
								display_update = 1;
								hold_button_counter = 0;
								back_from_settings = 1;
							}
							
							if (hold_button_counter < 20) hold_button_counter+=1;								
						break;
						
					case BUTTON_RELEASE:
						if (to_settings == 0)
						{
							if (menu_item_cursor != 4)
							{
								if (menu_item_selected == 0) menu_item_selected = menu_item_cursor+1; else menu_item_selected = 0;	
							}
							else
							{	
								if (calibration_step != CALIBRATION_2)
								{
									active_view = CALIBRATION_VIEW;
									display_update = 1;	
								}
								else
								{
									calibration_step = CALIBRATION_1;
								}
							}
						}							
						else
						{
							to_settings = 0;
						}							
							
						if (hold_button_counter < 20)
						{
							display_update = 1;
						}
						hold_button_counter = 0;
					break;
				}						
					
				update_flag = 0;	
			break;
			
			case CALIBRATION_VIEW:			
				switch (buttons_status[BUTTON_PLUS])
				{
					case BUTTON_PRESS:
						switch (calibration_step)
						{
							case CALIBRATION_1:
								if (cal_const1 < 1000) cal_const1 += 1;
							break;
							
							case CALIBRATION_2:
								if (cal_const2 < 10000) cal_const2 += 1;
							break;
						}
						display_update = 1;
					break;
				}					
				
				switch (buttons_status[BUTTON_MINUS])
				{
					case BUTTON_PRESS:
						switch (calibration_step)
						{
							case CALIBRATION_1:
								if (cal_const1 > 0) cal_const1 -= 1;
							break;
							
							case CALIBRATION_2:
								if (cal_const2 > 0) cal_const2 -= 1;
							break;
						}
						display_update = 1;
					break;
				}
				
				switch (buttons_status[BUTTON_MENU])
				{
					case BUTTON_PRESS:
						switch (calibration_step)
						{
							case 0:
								calibration_step = CALIBRATION_2;
								display_update = 1;
							break;
							
							case 1:
								current_control_const_a = ((int32_t)100*((int32_t)cal_const2-(int32_t)cal_const1))/((int32_t)100-(int32_t)10);
								current_control_const_b = ((int32_t)100*(int32_t)cal_const2) - current_control_const_a*(int32_t)100;
								write_cal_eeprom();
								active_view = SETTINGS_VIEW;
								display_update = 1;
							break;
						}

					break;
				}
				update_flag = 0;
			break;
		}
		
		vt_start(VTIMER_2, VT2_PERIOD);
}

void auto_power_off()
{
	vt_clear_flag(VTIMER_3);
	
	if (auto_power_off_counter == auto_power_off_counter_max)	
	{
		PORTD &= ~(1<<PD6); 
	}		
	else 
	{
		auto_power_off_counter++;
	}		
	
	vt_start(VTIMER_3, VT3_PERIOD);
}

void reset_auto_off_counter()
{
	auto_power_off_counter = 0;
}

void piezzo_togle()
{
	vt_clear_flag(VTIMER_4);

	if (speaker_flag)
	{
		PORTD |= _BV(PIEZZO);
		speaker_flag = 0;
	}
	else
	{
		PORTD &= ~_BV(PIEZZO);
		speaker_flag = 1;
	}

	if (beep_lenght_counter++ < BEEP_LENGHT)
	{
		vt_start(VTIMER_4, VT4_PERIOD);
	}
	else
	{
		beep_lenght_counter = 0;
		PORTD &= ~_BV(PIEZZO);
	}
}

void menu_backlight_control()
{	
	switch (config_backlight)
	{
		case 0:	PORTD &= ~(1<<BACKLIGHT_CONTROL); break;
		case 1:	PORTD |= (1<<BACKLIGHT_CONTROL); break;
	}				
}

void menu_auto_off_control()
{
	switch (config_auto_off)
	{
		case 0:	auto_power_off_counter_max = 30; break;
		case 1: auto_power_off_counter_max = 60; break;
		case 2: auto_power_off_counter_max = 120; break;
		case 3: auto_power_off_counter_max = 300; break;
		case 4: auto_power_off_counter_max = 600; break;
	}
	
	display_update = 1;
}

void beep()
{
	if (config_beep) vt_start(VTIMER_4, VT4_PERIOD);	
}

void resistor_update()
{
	uint32_t temp_res;
	
	if ((LED_power_supply > LED_voltage)&(LED_current != 0))
	{
		temp_res = (((uint32_t)LED_power_supply-(uint32_t)LED_voltage)*(uint32_t)10000)/(uint32_t)LED_current;
		
		if (temp_res < 100000)
		{
			LED_resistor_unit = 0;
		}
		
		if ((temp_res >= 100000) & (temp_res < 100000000))
		{
			temp_res /= 1000;
			LED_resistor_unit = 10;
		}
		
		if (temp_res >= 100000000)
		{
			temp_res /= 1000000;
			LED_resistor_unit = 20;
		}
		
		if (temp_res < 1000)
		{
			LED_resistor = temp_res;
			LED_resistor_unit += 0;
		}
		
		if ((temp_res >= 1000) & (temp_res < 10000))
		{
			LED_resistor = temp_res/10;
			LED_resistor_unit += 1;
		}
		
		if (temp_res >= 10000)
		{
			LED_resistor = temp_res/100;
			LED_resistor_unit += 2;
		}
		
	}
	else
	{
		LED_resistor = 0;
		LED_resistor_unit = 0;
	}
}

void read_config_eeprom()
{
	config_backlight = eeprom_read_byte(0x01);
	config_beep = eeprom_read_byte(0x02);
	config_auto_off = eeprom_read_byte(0x03);
	config_resistor = eeprom_read_byte(0x04);
}

void write_config_eeprom()
{
	eeprom_write_byte(0x01, config_backlight);
	eeprom_write_byte(0x02, config_beep);	
	eeprom_write_byte(0x03, config_auto_off);
	eeprom_write_byte(0x04, config_resistor);
	
	if (eeprom_read_byte(0x00) != 55) eeprom_write_byte(0x00, 55);
}

void read_cal_eeprom()
{
	current_control_const_a = eeprom_read_dword(0x06);
	current_control_const_b = eeprom_read_dword(0x0A);
}

void write_cal_eeprom()
{
	eeprom_write_dword(0x06, current_control_const_a);
	eeprom_write_dword(0x0A, current_control_const_b);
	
	if (eeprom_read_byte(0x05) != 55) eeprom_write_byte(0x05, 55);
}

int main(void)
{
	// Init I/0, Timers, ADC, ...
	peripherals_init();
	
	// Read configuration from eeprom memory or use default values
	if (eeprom_read_byte(0x00) == 55) read_config_eeprom();
	
	menu_backlight_control();
	menu_auto_off_control();
	
	// Read calibration values from eeprom memory or use default values
	if (eeprom_read_byte(0x05) == 55) read_cal_eeprom();
	
	// Buttons init
	for (uint8_t i=0; i<NUMBER_OF_BUTTONS; i++)
	{
		buttons_status[i] = BUTTON_NONE;	
	}
	
	// Initialize LCD
	nokia_3310_lcd_init();
	nokia_3310_lcd_clear();
	
	// Display welcome screen
	show_welcome_screen();
	nokia_3310_lcd_clear();
	
	// Global Interrupt Enable
	sei();
	
	// Start all virtual timers
	vt_start(VTIMER_0, VT0_PERIOD);
	vt_start(VTIMER_1, VT1_PERIOD);
	vt_start(VTIMER_2, VT2_PERIOD);
	vt_start(VTIMER_3, VT3_PERIOD);
	vt_start(VTIMER_4, VT4_PERIOD);
 
    while(1)
    {
		if (timer_flags & (1<<VTIMER_0)) refresh_display();
		if (timer_flags & (1<<VTIMER_1)) button_scan();
		if (timer_flags & (1<<VTIMER_2)) control_loop();
		if (timer_flags & (1<<VTIMER_3)) auto_power_off();
		if (timer_flags & (1<<VTIMER_4)) piezzo_togle();
    }
}