/* LCD library for noka 3110 display
 * Date: 27 Oct 2009
 * Author: A Lindsay (http://blog.thiseldo.co.uk/?p=383)
 * Credits: based on original nuelectronics library extensively tidied up
 * and with functions re-named and recoded.
 * Additional graphics code added with options to disnable graphics if
 * not required, see nokia_3110_lcd.h
 *
 * Updated: 10th November by John Crouchley
 * 1) Allow for multiple SPI devices by
 *    a) allowing SPI_CS to reside on any port
 *    b) set SPCR as part of LCDENABLE
 * 2) move get_key into the class
 * 3) Fix bug in writeChar caused column 6 of any character not to be updated
 * 4) add code to ensure values for cursor_row and cursor_col are kept within array bounds
 * 5) Inherit Nokia_3310_lcd class from Print and add a write function
 *    This enables all normal print behaviour, including numbers
 *
 * Updated: 14th November by Andrew Lindsay
 * 1) Create new big font that is cleaner.
 *
 * Updated: 9th December 2009 by John Crouchley
 * 1) PORTB2 (Arduino digital 10) must be set to OUTPUT and HIGH whilst SPCR is being changed
 *    Failure to do this can lead to intermittent SPI errors.
 *    If there is another SPI device already using PORTB2 then it will naturally be high already
 *    Warning - if you use this pin for anything else (other than another SPI device enable)
 *    then you may have problems.
 *
 * Updated: 26th January 2010 by John Crouchley
 * 1) Initialisation values (contrast and temp coefficient) updated for later LCDs (got a black screen)
 *    values taken from updated driver on Nuelectronics site. This still works with the earlier lcds.
 *
 * Update: 9th July 2011 by Miroslav Batek
 * 1) Conversion to C
 * 2) Removed joystick part
 * 3) Pin assigment modified
 *
 */

#include <avr/pgmspace.h>
#include <avr/io.h>
#include <util/delay.h>
#include "nokia_3310_lcd.h"
#include "font_5x7.h"
#include "font_big.h"
#include "diode.h"

// defines for accessing lcd
#define SPI_INIT	SPCR = 0x51
#define LCDENABLE	SPI_INIT; SPI_CS_PORT &= ~(1<<SPI_CS)	// Enable LCD
#define LCDDISABLE	SPI_CS_PORT |= (1<<SPI_CS)	// disable LCD
#define LCDCMDMODE	PORTB &= ~(1<<LCD_DC)	// Set LCD in command mode
#define LCDDATAMODE	PORTB |= (1<<LCD_DC)	// Set LCD to Data mode

#define LCDCOLMAX	84
#define LCDROWMAX	6
#define LCDPIXELROWMAX	48

// current cursor postition
static unsigned char cursor_row = 0; /* 0-5 */
static unsigned char cursor_col = 0; /* 0-83 */

#ifdef USE_GRAPHIC
static unsigned char lcd_buffer[LCDROWMAX][LCDCOLMAX];
#endif

/*
 * Name         : init 
 * Description  : Main LCD initialisation function
 * Argument(s)  : None 
 * Return value : None 
 */
void nokia_3310_lcd_init(void){
    DDRB |= (1<<LCD_DC)|(1<<LCD_RST)|(1<<SPI_MOSI)|(1<<SPI_SCK);
    DDRD |= (1<<LCD_BL);
    SPI_CS_DDR |= (1<<SPI_CS);
     
    LCDDISABLE;
    PORTB &= ~(1<<LCD_RST);
  	
    _delay_us(1);

    PORTB |= (1<<LCD_RST);  
    SPI_INIT;   			// enable SPI master, fosc/16 = 1MH
    PORTD |= (1<<LCD_BL);  	// turn on backlight

    nokia_3310_lcd_writeCommand(0x21);		// LCD Extended Commands
    //writeCommand(0xe0); 	// Set LCD Vop (Contrast)
    nokia_3310_lcd_writeCommand(0xC0); 	// Set LCD Vop (Contrast) **NEW**
    //writeCommand(0x04);  	// Set temp coefficient
    nokia_3310_lcd_writeCommand(0x06);  	// Set temp coefficient   **NEW**
    nokia_3310_lcd_writeCommand(0x13);		// LCD bias mode1:48
    nokia_3310_lcd_writeCommand(0x20);		// LCD Standard Commands, Horizontal addressing mode
    nokia_3310_lcd_writeCommand(0x0c);		// LCD in normal mode
    nokia_3310_lcd_clear();
}


/*
 * Name         : writeCommand 
 * Description  : Sends command to display controller 
 * Argument(s)  : command - The command to be sent 
 * Return value : none 
 */
void nokia_3310_lcd_writeCommand(unsigned char command) {
    LCDENABLE;
    LCDCMDMODE;
    SPDR = command;		// Send data to display controller

    while (!(SPSR & 0x80));   	// Wait until Tx register empty

    LCDDISABLE;
}


/*
 * Name         : writeData 
 * Description  : Sends data to the display controller 
 * Argument(s)  : Date - Data to be sent 
 * Return value : none 
 */
void nokia_3310_lcd_writeData(unsigned char data) {
    LCDENABLE;
    LCDDATAMODE;
    SPDR = data;		// Send data	

    while (!(SPSR & 0x80));	// Wait until Tx register empty   

    LCDDISABLE;
}


#ifdef USE_BITMAP
/*
 * Name         : drawBitmap
 * Description  : Sends a bitmap image stored in ram to the display 
 * Argument(s)  : x, y - Position on screen, x 0-83, y 1-6
 *                map - pointer to data
 *                size_x,size_y - Size of the image in pixels,
 *                size_y is multiple of 8
 * Return value : none 
 */
void nokia_3310_lcd_drawBitmap( unsigned char x,unsigned char y, unsigned char *map, unsigned char size_x,unsigned char size_y) {

    unsigned int i,n;
    unsigned char row;
    
    row = (size_y % 8 == 0 ) ? size_y / 8 : size_y / 8 + 1;
//    if (size_y % 8==0) {
//	row = size_y/8;  
//    } else {
//	row = size_y/8+1;
//    }
    
    for (n=0;n<row;n++) {
      	nokia_3310_lcd_gotoXY(x,y);
        for(i=0; i<size_x; i++) {
            nokia_3310_lcd_writeData(map[i+n*size_x]);
          }
        y++;
      }
}


/*
 * Name         : drawBitmapP 
 * Description  : Sends a bitmap image stored in progmen/flash to the display 
 *                Use Bitmap2LCD to create bitmap files.
 *                Settings: Full screen Sixe 84 x 48,
 *                Vertical Downwards, MSB Last
 *                Use comma seperaters
 *                Change .h file definition to be:
 *                	static const prog_char  imageName [] PROGMEM = {
 * 		  
 * Argument(s)  : x, y - Position on screen, x 0-83, y 1-6
 *                map - pointer to data
 *                size_x,size_y - Size of the image in pixels,
 *                size_y is multiple of 8
 * Return value : none 
 */
void nokia_3310_lcd_drawBitmapP( unsigned char x,unsigned char y, const prog_char *map, unsigned char size_x,unsigned char size_y)
  {
    unsigned int i,n;
    unsigned char row;
    const prog_char *dptr;
    
    row = (size_y % 8 == 0 ) ? size_y / 8 : size_y / 8 + 1;
//    if (size_y % 8 ==0)
//    	row=size_y/8; 
//    else
//    	row=size_y/8+1;
    
    for (n=0;n<row;n++) {
      	nokia_3310_lcd_gotoXY(x,y);
        for(i=0; i<size_x; i++) {
            dptr = map + (i + n * size_x);
            nokia_3310_lcd_writeData( pgm_read_byte( dptr ) );
          }
        y++;                       
    }      
}


/*
 * Name         : clearBitmap 
 * Description  : Clear an area of the screen, usually to blank out a
 * 		  previously drawn image or part of image.
 * Argument(s)  : x, y - Position on screen, x 0-83, y 1-6
 *                size_x,size_y - Size of the image in pixels,
 *                size_y is multiple of 8
 * Return value : none 
 */
void nokia_3310_lcd_clearBitmap( unsigned char x,unsigned char y, unsigned char size_x,unsigned char size_y)
  {
    unsigned int i,n;
    unsigned char row;
    
    row = (size_y % 8 == 0 ) ? size_y / 8 : size_y / 8 + 1;
//    if (size_y % 8==0)
//	    row=size_y/8;  
//    else
//    	    row=size_y/8+1;
    
    for (n=0;n<row;n++) {
      	nokia_3310_lcd_gotoXY(x,y);
        for(i=0; i<size_x; i++) {
            nokia_3310_lcd_writeData( 0x00 );
        }
        y++;                       
    }      
}
#endif


/*
 * Name         : writeString 
 * Description  : Write a string to the LCD from current position
 * Argument(s)  : x - Column to start at, 0-83
 * 		  y - Row to start, 0-6
 * 		  s - Pointer to start of string
 * 		  mode - Mode of string, reverse or normal 
 *                 
 * Return value : none 
 */
void nokia_3310_lcd_writeString(unsigned char x,unsigned char y,char *string, char mode){
    nokia_3310_lcd_gotoXY(x,y);
    while (*string)
		nokia_3310_lcd_writeChar(*string++, mode);
}


/*
 * Name         : writeStringP
 * Description  : Write a string stored in PROGMEM to the LCD
 * 		  from current position
 * Argument(s)  : x - Column to start at, 0-83
 * 		  y - Row to start, 0-6
 * 		  s - Pointer to start of string
 * 		  mode - Mode of string, reverse or normal 
 * Return value : none 
 */
void nokia_3310_lcd_writeStringP ( unsigned char x,unsigned char y,const char *string, char mode )
{
    nokia_3310_lcd_gotoXY(x,y);
    while ( *string )
	nokia_3310_lcd_writeChar( *string++, mode );
}


/*
 * Name         : writeStringBig 
 * Description  : Write a string using big font to position x,y
 * 		  Note: bigfont only includes digits, '.', '+', '-'
 * Argument(s)  : x,y - starting position on screen, x=0-83, y=0-6 
 *                string - string pointer of data to display
 *                mode - reverse or normal
 * Return value : none 
 */
void nokia_3310_lcd_writeStringBig ( unsigned char x,unsigned char y,
		char *string, char mode ){
    while ( *string ){
        nokia_3310_lcd_writeCharBig( x, y, *string , mode );
        
        // For decimal point use 5 pixel gap instead of 12 to not
	// make spacing look odd
        if(*string++ == '.')
          x += 5;
        else
          x += 12;
    }	
}


/*
 * Name         : writeCharBig 
 * Description  : Write a single big character to screen 
 * 		  Note: bigfont only includes digits, '.', '+', '-'
 * Argument(s)  : x,y - starting position on screen, x=0-83, y=0-6 
 *                ch - character to display
 *                mode - reverse or normal
 * Return value : none
 */
void nokia_3310_lcd_writeCharBig (unsigned char x,unsigned char y,
		unsigned char ch, char mode) {
unsigned char i, j;
unsigned char *pFont;
unsigned char ch_dat;
unsigned char colsUsed = 12;

   pFont = (unsigned char *) big_number;
   
   if(ch == '.')
   {
     ch = 10;
     colsUsed=5;
   }
   else if (ch == '+')
     ch = 11;
   else if (ch == '-')
     ch = 12;
   else
     ch = ch & 0x0f;

#ifdef USE_GRAPHIC
   if (x > LCDCOLMAX - colsUsed) x = LCDCOLMAX - colsUsed ; // ensure space is available for the character
   if (y > LCDROWMAX - 3) y = LCDROWMAX - 3 ; // ensure space is available for the character
#endif

   for(i=0;i<3;i++) {
     nokia_3310_lcd_gotoXY( x, y+i);
 
     for(j=0; j<colsUsed; j++) {
       ch_dat =  pgm_read_byte(pFont+ch*48 + i*16 +j);	// 16 cols in data for char even if we use fewer.
       //ch_dat =  pgm_read_byte(pFont+ch*48 + i*12 +j);	// 12 cols in data for char even if we use fewer.
#ifdef USE_GRAPHIC
       lcd_buffer[cursor_row][cursor_col + j] = (mode == MENU_NORMAL)? ch_dat : (ch_dat^0xff);
#endif
       nokia_3310_lcd_writeData( (mode == MENU_NORMAL)? ch_dat : (ch_dat^0xff));
     }
   } 
}


/*
 * Name         : writeChar 
 * Description  : Write a single normal font character to screen
 * 		  at current cursor position
 * Argument(s)  : ch - character to display
 *                mode - reverse or normal
 * Return value : none
 */
void nokia_3310_lcd_writeChar(unsigned char ch, char mode) {
	unsigned char j;
  
#ifdef USE_GRAPHIC
      if (cursor_col > LCDCOLMAX - 6) cursor_col = LCDCOLMAX - 6; // ensure space is available for the character
      if (cursor_row > LCDROWMAX - 1) cursor_row = LCDROWMAX - 1; // ensure space is available for the character
	lcd_buffer[cursor_row][cursor_col] = 0x00;
	for(j=0; j<5; j++) {
		lcd_buffer[cursor_row][cursor_col + j] =  pgm_read_byte(&(smallFont [(ch-32)*5 + j] ));
        }

	lcd_buffer[cursor_row][cursor_col + 5] = 0x00;

	for(j=0; j< 6; j++) {
		if( mode == MENU_NORMAL ) 
			nokia_3310_lcd_writeData(lcd_buffer[cursor_row][cursor_col++]);
		else
			nokia_3310_lcd_writeData(lcd_buffer[cursor_row][cursor_col++] ^ 0xff);
		if (cursor_col >= LCDCOLMAX)
		{
			cursor_col=0;
			cursor_row++;
			if (cursor_row >= LCDROWMAX) cursor_row=0;
		}
	}
#else
	for(j=0; j<5; j++) {
		if( mode == MENU_NORMAL ) 
			nokia_3310_lcd_writeData( pgm_read_byte(&(smallFont [(ch-32)*5 + j] )) );
		else
			nokia_3310_lcd_writeData( pgm_read_byte(&(smallFont [(ch-32)*5 + j] )) ^ 0xff );
	}
	nokia_3310_lcd_writeData( (mode == MENU_NORMAL) ? 0x00 : 0xff );
#endif
}

/*
 * Name         : write
 * Description  : write a character - override function in Print class
 * Argument(s)  : character to write
 * Return value : none
 */
void nokia_3310_lcd_write(uint8_t b)
{
	nokia_3310_lcd_writeChar(b,0);
}

/*
 * Name         : nokia_3310_lcd_gotoXY
 * Description  : Move text position cursor to specified position
 * Argument(s)  : x, y - Position, x = 0-83, y = 0-6
 * Return value : none
 */
void nokia_3310_lcd_gotoXY(unsigned char x, unsigned char y) {
      if (x > LCDCOLMAX - 1) x = LCDCOLMAX - 1 ; // ensure within limits
      if (y > LCDROWMAX - 1) y = LCDROWMAX - 1 ; // ensure within limits

	nokia_3310_lcd_writeCommand (0x80 | x);   //column
	nokia_3310_lcd_writeCommand (0x40 | y);   //row

	cursor_row = y;
	cursor_col = x;
}


/*
 * Name         : clear
 * Description  : Clear the screen and display buffer 
 * Argument(s)  : none
 * Return value : none
 */
void nokia_3310_lcd_clear(void) {
	int i,j;
	
	nokia_3310_lcd_gotoXY (0,0);  	//start with (0,0) home position

	for(i=0; i< LCDROWMAX; i++) {
		for(j=0; j< LCDCOLMAX; j++) {
			nokia_3310_lcd_writeData( 0x00 );
#ifdef USE_GRAPHIC
			lcd_buffer[i][j] = 0x00;
#endif
		}
	}
   
      nokia_3310_lcd_gotoXY (0,0);	//bring the XY position back to (0,0)
}


/*
 * Name         : update
 * Description  : Write the screen buffer to the display memory
 * Argument(s)  : none
 * Return value : none
 */
void nokia_3310_lcd_update(void) {
#ifdef USE_GRAPHIC
	int i,j;

	for(i=0; i< LCDROWMAX; i++) {
		nokia_3310_lcd_gotoXY (0,i);
		for(j=0; j< LCDCOLMAX; j++) {
			nokia_3310_lcd_writeData(lcd_buffer[i][j]);
		}
	}
#endif 
	nokia_3310_lcd_gotoXY (0,0);	//bring the XY position back to (0,0)
}


#ifdef USE_GRAPHIC
/*
 * Name         : setPixel
 * Description  : Set a single pixel either on or off, update display buffer.
 * Argument(s)  : x,y - position, x = 0-83, y = 0-6
 *                c - colour, either PIXEL_ON, PIXEL_OFF or PIXEL_XOR
 * Return value : none
 */
void nokia_3310_lcd_setPixel( unsigned char x, unsigned char y, unsigned char c ) {
unsigned char value;
unsigned char row;
	
	if( x < 0 || x >= LCDCOLMAX || y < 0 || y >= LCDPIXELROWMAX ) return;

	row = y / 8;

	value = lcd_buffer[row][x];
	if( c == PIXEL_ON ) {
		value |= (1 << (y % 8));
	} else if( c == PIXEL_XOR ) {
		value ^= (1 << (y % 8));
	} else {
		value &= ~(1 << (y % 8));
	}

	lcd_buffer[row][x] = value;

	nokia_3310_lcd_gotoXY (x,row);
	nokia_3310_lcd_writeData(value);
}


/*
 * Name         : drawLine
 * Description  : Draws a line between two points on the display.
 * Argument(s)  : x1, y1 - Absolute pixel coordinates for line origin.
 *                x2, y2 - Absolute pixel coordinates for line end.
 *                c - either PIXEL_ON, PIXEL_OFF or PIXEL_XOR
 * Return value : none
 */
void nokia_3310_lcd_drawLine(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char c) {
    int dx, dy, stepx, stepy, fraction;

    /* Calculate differential form */
    /* dy   y2 - y1 */
    /* -- = ------- */
    /* dx   x2 - x1 */

    /* Take differences */
    dy = y2 - y1;
    dx = x2 - x1;

    /* dy is negative */
    if ( dy < 0 ) {
        dy    = -dy;
        stepy = -1;
    } else {
        stepy = 1;
    }

    /* dx is negative */
    if ( dx < 0 ) {
        dx    = -dx;
        stepx = -1;
    } else {
        stepx = 1;
    }

    dx <<= 1;
    dy <<= 1;

    /* Draw initial position */
    nokia_3310_lcd_setPixel( x1, y1, c );

    /* Draw next positions until end */
    if ( dx > dy ) {
        /* Take fraction */
        fraction = dy - ( dx >> 1);
        while ( x1 != x2 ) {
            if ( fraction >= 0 ) {
                y1 += stepy;
                fraction -= dx;
            }
            x1 += stepx;
            fraction += dy;

            /* Draw calculated point */
            nokia_3310_lcd_setPixel( x1, y1, c );
        }
    } else {
        /* Take fraction */
        fraction = dx - ( dy >> 1);
        while ( y1 != y2 ) {
            if ( fraction >= 0 ) {
                x1 += stepx;
                fraction -= dy;
            }
            y1 += stepy;
            fraction += dx;

            /* Draw calculated point */
            nokia_3310_lcd_setPixel( x1, y1, c );
        }
    }
}


/*
 * Name         : drawRectangle
 * Description  : Draw a rectangle given to top left and bottom right points
 * Argument(s)  : x1, y1 - Absolute pixel coordinates for top left corner
 *                x2, y2 - Absolute pixel coordinates for bottom right corner
 *                c - either PIXEL_ON, PIXEL_OFF or PIXEL_XOR
 * Return value : none
 */
void nokia_3310_lcd_drawRectangle(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char c){
	nokia_3310_lcd_drawLine( x1, y1, x2, y1, c );
	nokia_3310_lcd_drawLine( x1, y1, x1, y2, c );
	nokia_3310_lcd_drawLine( x1, y2, x2, y2, c );
	nokia_3310_lcd_drawLine( x2, y1, x2, y2, c );
}


/*
 * Name         : drawFilledRectangle
 * Description  : Draw a filled rectangle given to top left and bottom right points
 * 		  just simply draws horizontal lines where the rectangle would be
 * Argument(s)  : x1, y1 - Absolute pixel coordinates for top left corner
 *                x2, y2 - Absolute pixel coordinates for bottom right corner
 *                c - either PIXEL_ON, PIXEL_OFF or PIXEL_XOR
 * Return value : none
 */
void nokia_3310_lcd_drawFilledRectangle(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char c) {
	for(int i=y1; i <= y2; i++ ) {
		nokia_3310_lcd_drawLine( x1, i, x2, i, c );
	}
}


/*
 * Name         : drawCircle
 * Description  : Draw a circle using Bresenham's algorithm. 
 * 		  Some small circles will look like squares!!
 * Argument(s)  : xc, yc - Centre of circle
 * 		  r - Radius
 * 		  c - either PIXEL_ON, PIXEL_OFF or PIXEL_XOR
 * Return value : None 
 */
void nokia_3310_lcd_drawCircle(unsigned char xc, unsigned char yc, unsigned char r, unsigned char c) {
	int x=0;
	int y=r;
	int p=3-(2*r);

        nokia_3310_lcd_setPixel( (uint8_t)(xc+x),(uint8_t)(yc-y), c);

	for(x=0;x<=y;x++) {
		if (p<0) {
			y=y;
			p=(p+(4*x)+6);
		} else {
			y=y-1;
			p=p+((4*(x-y)+10));
		}

		nokia_3310_lcd_setPixel((uint8_t)(xc+x),(uint8_t)(yc-y), c);
		nokia_3310_lcd_setPixel((uint8_t)(xc-x),(uint8_t)(yc-y), c);
		nokia_3310_lcd_setPixel((uint8_t)(xc+x),(uint8_t)(yc+y), c);
		nokia_3310_lcd_setPixel((uint8_t)(xc-x),(uint8_t)(yc+y), c);
		nokia_3310_lcd_setPixel((uint8_t)(xc+y),(uint8_t)(yc-x), c);
		nokia_3310_lcd_setPixel((uint8_t)(xc-y),(uint8_t)(yc-x), c);
		nokia_3310_lcd_setPixel((uint8_t)(xc+y),(uint8_t)(yc+x), c);
		nokia_3310_lcd_setPixel((uint8_t)(xc-y),(uint8_t)(yc+x), c);
	}
}
#endif
