/* Virtual Timers library
 * Date: 24 Jul 2011
 * Author: Miroslav Batek <diy4fun@gmail.com>  http://diy4fun.blogspot.com
 * Credits: based on article from Circuit Cellar
 */

/* Example:
 * 
 * #define VTIMER0 0
 * #define VTIMER1 1
 *
 * #define VT0_PERIOD 1000
 * #define VT1_PERIOD 2
 *
 * int main(void){
 *  vt_start(VTIMER1, VTIMER1_PERIOD);
 *  vt_start(VTIMER2, VTIMER2_PERIOD);
 *
 *  while (1){
 *   if (timer_flags & _BV(VTIMER0)) led_blink();
 *	 if (timer_flags & _BV(VTIMER1)) input_scan();
 *  }
 * }
 *
 */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>

volatile uint8_t timer_flags;			// virtual timer flags
unsigned int timer_counters[8];			// counters for virtual timers

/*
 * Name         : vt_interrupt 
 * Description  : Call periodically in timer interrupt routine to define main clock for virtual timers
 * Argument(s)  : None 
 * Return value : None 
 */
void vt_interrupt(void)
{
	unsigned int use;
	uint8_t i;

	for (i = 0; i < 8; i++)
	{
		use = timer_counters[i];
		if (use) 
		{
			timer_counters[i] = --use; 
			if (!use) timer_flags |= _BV(i);
		}
	}
}

/*
 * Name         : vt_start 
 * Description  : Start virtual timer
 * Argument(s)  : who ... the timer number we want to use (0-7)
 *				  count ... number of ticks to run it (period)
 * Return value : None 
 */
void vt_start(uint8_t who, unsigned int count)
{
	uint8_t mask;

	who &= 0x07;
	mask = _BV(who);
	cli();
	timer_counters[who] = count;
	if (count) timer_flags &= ~mask; else timer_flags |= mask;
	sei();
}

/*
 * Name         : vt_clear_flag
 * Description  : Start virtual timer
 * Argument(s)  : who ... the timer number we want to use (0-7)
 * Return value : None 
 */
void vt_clear_flag(uint8_t who)
{
	who &= 0x07;
	cli();
	timer_flags &= ~_BV(who);
	sei();
}
